# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-08-24 14:45
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0023_auto_20160819_1310'),
    ]

    operations = [
        migrations.AlterField(
            model_name='change',
            name='at',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



# Imports from other dependencies.  # NOQA
from psycopg2.extras import DateRange
from psycopg2.extras import DateTimeTZRange
from rest_framework import serializers
from rest_framework.serializers import ModelSerializer


# Imports from budget.
from budget.models import Change
from budget.models import ContentPlacement
from budget.models import Headline
from budget.models import HeadlineVote
from budget.models import Item
from budget.models import Package
from budget.models import PrintPublication
# from budget.models import PrintSection


class UserMetadataMixin(object):
    """Mixin that adds an audit trail to REST operations.

    On change, creates a ChangeRecord module with user and model info.
    On create, adds user info to model before save. Base class should
    implement _get_package(), which will be used to get related package
    for ChangeRecord.
    """
    def _get_user(self):
        """Get the user making the request."""
        user = None
        request = self.context.get('request')
        if request and hasattr(request, 'user'):
            user = request.user
        return user

    def _get_package(self, instance):
        """Get package instance for any models able to connect to them.

        Works on any models that can connect to a package (even those
        that connect indirectly). Serializers should override this to
        implement model-specific logic.
        """
        return None

    def create(self, validated_data):
        """Fill in model's created_by field with appropriate value.

        Sets this field value to the requesting user's primary key.
        """
        validated_data['created_by'] = self._get_user().id
        return super(UserMetadataMixin, self).create(validated_data)

    def update(self, instance, validated_data):
        """Model-update behavior.

        On update, let REST Framework do its usual saving for the
        affected instance. Then save a change record once that update
        has completed.
        """
        instance = super(
            UserMetadataMixin, self).update(instance, validated_data)

        Change.objects.create(
            by=self._get_user().id,
            item_id=instance.pk,
            item_content_type=instance._meta.model_name,
            package=self._get_package(instance)
        )
        return instance


class EmptyArrayJsonField(serializers.JSONField):
    """Field serializer for JSONFields.

    Returns an empty array when the JSONField is null, giving API
    consumers a more reliable data type.
    """
    def get_attribute(self, instance):
        attr = super(EmptyArrayJsonField, self).get_attribute(instance)
        if attr is None:
            return []
        return attr


class DateRangeField(serializers.ListField):
    """Custom serializer for django.postgres.fields.DateRangeFields.

    Only supports storing ranges with defined upper and lower bounds.
    """
    child = serializers.DateField()
    range_class = DateRange

    def to_representation(self, obj):
        return [obj.lower, obj.upper]

    def to_internal_value(self, data):
        data = super(DateRangeField, self).to_internal_value(data)

        try:
            lower = data[0]
            upper = data[1]
        except IndexError:
            raise serializers.ValidationError(
                'Incorrect format. Expected an Array with two items.')

        if lower >= upper:
            raise serializers.ValidationError(
                'The upper date bound must be greater than the lower bound.')

        return self.range_class(lower=lower, upper=upper)


class DateTimeRangeField(DateRangeField):
    """Custom serializer for django.postgres.fields.DateTimeRangeFields.

    Works just like DateRangeField serializer.
    """
    child = serializers.DateTimeField()
    range_class = DateTimeTZRange


class ContentPlacementSerializer(UserMetadataMixin, ModelSerializer):
    package = serializers.PrimaryKeyRelatedField(
        required=True,
        many=False,
        queryset=Package.objects.all()
    )
    destination = serializers.PrimaryKeyRelatedField(
        required=True,
        many=False,
        queryset=PrintPublication.objects.all()
    )
    run_date = DateRangeField(allow_null=True, required=False)

    class Meta:  # NOQA
        model = ContentPlacement
        read_only_fields = (
            # 'slug',
            'created_by',
        )
        fields = (
            'id',
            'url',
            'package',  # ForeignKey.
            'destination',  # ForeignKey.
            'placement_types',  # ArrayField.
            'page_number',  # PositiveSmallIntegerField.
            'placement_details',  # CharField.
            'run_date',  # Date range.
            'external_slug',  # CharField.
            'is_finalized',  # Bool.
            'created_by',
        )
        extra_kwargs = {
            'url': {'view_name': 'budget:contentplacement-detail'},
        }


class HeadlineSerializer(UserMetadataMixin, ModelSerializer):
    def _get_package(self, instance):
        return instance.package

    class Meta:  # NOQA
        model = Headline
        fields = ('id', 'url', 'package', 'text', 'winner')
        extra_kwargs = {
            'url': {'view_name': 'budget:headline-detail'},
        }


class HeadlineVoteSerializer(UserMetadataMixin, ModelSerializer):
    def _get_package(self, instance):
        return instance.headline.package

    class Meta:  # NOQA
        model = HeadlineVote
        fields = ('id', 'url', 'headline', 'voter', 'timestamp')
        extra_kwargs = {
            'url': {'view_name': 'budget:headlinevote-detail'},
        }


class ItemSerializer(UserMetadataMixin, ModelSerializer):
    editors = EmptyArrayJsonField(allow_null=True, required=False)
    authors = EmptyArrayJsonField()

    slug = serializers.ReadOnlyField(source='full_slug')

    def _get_package(self, instance):
        if instance.primary_for_package:
            return instance.primary_for_package
        elif instance.additional_for_package:
            return instance.additional_for_package

    def validate_authors(self, value):
        """Ensure at least one author is provided"""
        if value is None or len(value) == 0:
            raise serializers.ValidationError(
                'At least one author is required.')
        return value

    def validate(self, attrs):
        """Run model validation that REST doesn't run automatically.

        Specifically, calls Item.full_clean() for validation since REST
        doesn't call it by default.
        """
        # Only validate if both primary and additional exist (it could be a
        # PATCH)
        if 'primary_for_package' in attrs \
                and 'additional_for_package' in attrs:
            if attrs.get('primary_for_package', None) is None and \
                    attrs['additional_for_package'] is None:
                err_msg = ('Items must be connected to a package as either '
                           'primary or additional content.')
                raise serializers.ValidationError({
                    'primary_for_package': err_msg,
                    'additional_for_package': err_msg
                })
            if attrs['primary_for_package'] is not None and \
                    attrs['additional_for_package'] is not None:
                err_msg = ('Items cannot be connected to a package as both '
                           'primary and additional content.')
                raise serializers.ValidationError({
                    'primary_for_package': err_msg,
                    'additional_for_package': err_msg
                })
        return super(ItemSerializer, self).validate(attrs)

    class Meta:  # NOQA
        model = Item
        read_only_fields = (
            'slug',
            'created_by',
        )
        fields = (
            'id',
            'url',
            'slug_key',
            'slug',
            'type',
            'editors',
            'authors',
            'is_ready',
            'budget_line',
            'length',
            'primary_for_package',
            'additional_for_package',
            'created_by',
        )
        extra_kwargs = {
            'url': {'view_name': 'budget:item-detail'},
        }


class PackageSerializer(UserMetadataMixin, ModelSerializer):
    publish_date = DateTimeRangeField()
    # print_run_date = DateRangeField(allow_null=True, required=False)

    primary_content = serializers.PrimaryKeyRelatedField(read_only=True)
    additional_content = serializers.PrimaryKeyRelatedField(
        required=False,
        many=True,
        queryset=Item.objects.all()
    )
    headline_candidates = serializers.PrimaryKeyRelatedField(
        required=False,
        many=True,
        queryset=Headline.objects.all()
    )
    # print_section = serializers.PrimaryKeyRelatedField(
    #     required=False,
    #     many=True,
    #     queryset=PrintSection.objects.all()
    # )

    slug = serializers.ReadOnlyField(source='full_slug')
    slug_date = serializers.ReadOnlyField(source='slugified_date')

    def _get_package(self, instance):
        return instance

    class Meta:  # NOQA
        model = Package
        read_only_fields = (
            'created_by',
            'slug',
        )
        fields = (
            'id',
            'url',
            'vertical',
            'hub',
            'slug',
            'slug_date',
            'slug_key',
            'pub_date',
            'pub_date_resolution',
            'publish_date',
            'publish_date_resolution',
            'published_url',
            'headline_status',
            'headline_candidates',
            # 'print_run_date',
            # 'print_system_slug',
            # 'print_placements',
            # 'print_section',
            # 'is_print_placement_finalized',
            'notes',
            'primary_content',
            'additional_content',
            'created_by',
        )
        extra_kwargs = {
            'url': {'view_name': 'budget:package-detail'},
        }


class PrintPublicationSerializer(ModelSerializer):
    class Meta:  # NOQA
        depth = 1
        model = PrintPublication
        fields = (
            'id',
            'url',
            'name',
            'slug',
            'priority',
            'is_active',
            'sections'
        )
        extra_kwargs = {
            'url': {'view_name': 'budget:printpublication-detail'},
        }



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-11-14 19:22
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0038_auto_20171028_1706'),
    ]

    operations = [
        migrations.AddField(
            model_name='contentplacement',
            name='page_number',
            field=models.PositiveSmallIntegerField(blank=True, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-27 16:20
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0013_auto_20160627_1006'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='item',
            name='subscribers',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-09-24 20:49
from __future__ import unicode_literals

import django.contrib.postgres.fields
import django.contrib.postgres.fields.ranges
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0032_auto_20170924_1204'),
    ]

    operations = [
        migrations.CreateModel(
            name='ContentPlacement',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_by', models.PositiveSmallIntegerField(db_index=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('placement_types', django.contrib.postgres.fields.ArrayField(base_field=models.CharField(blank=True, max_length=100, null=True), blank=True, null=True, size=None)),
                ('placement_details', models.CharField(help_text='E.G., print page number.', max_length=25)),
                ('run_date', django.contrib.postgres.fields.ranges.DateRangeField(blank=True, null=True)),
                ('external_slug', models.CharField(blank=True, max_length=250, null=True)),
                ('is_finalized', models.BooleanField(default=False)),
                ('destination', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='placed_content', to='budget.PrintPublication')),
                ('package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='placements', to='budget.Package')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-05-31 22:48
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0006_auto_20160531_1717'),
    ]

    operations = [
        migrations.AlterField(
            model_name='printpublication',
            name='is_active',
            field=models.BooleanField(default=True, help_text=b'Should this publication be included in the current options users can choose from? If false, this publication will be archived, but it may be re-activated in the future.', verbose_name=b'Active?'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.6 on 2016-06-23 20:21
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0008_package_publish_date'),
    ]

    operations = [
        migrations.CreateModel(
            name='PrintSection',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
                ('slug', models.SlugField(max_length=100)),
                ('priority', models.PositiveSmallIntegerField(default=10, help_text=b'In the list of all active print publications, how high should this entry show up?')),
                ('is_active', models.BooleanField(default=True, help_text=b'Should this publication be included in the current options users can choose from? If false, this publication will be archived, but it may be re-activated in the future.', verbose_name=b'Active?')),
                ('publication', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sections', to='budget.PrintPublication')),
            ],
            options={
                'ordering': ['publication__priority', 'publication__is_active', 'publication__slug', 'priority', 'is_active', 'slug'],
            },
        ),
        migrations.RemoveField(
            model_name='package',
            name='print_publication',
        ),
        migrations.AddField(
            model_name='package',
            name='print_section',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='budget.PrintSection'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-05-31 15:58
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0004_auto_20160424_1343'),
    ]

    operations = [
        migrations.AlterField(
            model_name='item',
            name='type',
            field=models.CharField(choices=[(b'text', b'Text'), (b'photo', b'Photo'), (b'video', b'Video'), (b'audio', b'Audio'), (b'graphic', b'Graphic'), (b'interactive', b'Interactive'), (b'aggregation', b'Aggregation'), (b'other', b'Other')], default=b'text', max_length=25),
        ),
    ]



# Imports from python.  # NOQA
import os
from setuptools import find_packages
from setuptools import setup


REPO_URL = 'https://github.com/DallasMorningNews/django-budget/'

PYPI_VERSION = '0.8.2'


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-budget',
    version=PYPI_VERSION,
    packages=find_packages(),
    # packages=find_packages(exclude=['demo']),
    include_package_data=True,
    license='AGPLv3',
    description=' '.join([
        'A tool to help newsrooms manage their content,',
        'from pitch to planning to production.'
    ]),
    long_description=README,
    url=REPO_URL,
    download_url='{repo_url}archive/{version}.tar.gz'.format(**{
        'repo_url': REPO_URL,
        'version': PYPI_VERSION,
    }),
    author='Allan James Vestal, The Dallas Morning News',
    author_email='newsapps@dallasnews.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        # Replace these appropriately if you are stuck on Python 2.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'Django>=1.9.0,<2.0',
        'django-admin-sortable2~=0.6.16',
        'django-editorial-staff>=0.5.2',
        'django-filter~=1.0.0',
        'djangorestframework~=3.6.0',
        'djangorestframework-camel-case~=0.2.0',
        'psycopg2~=2.6.1',
        'python-dateutil~=2.6.1',
        'pytz~=2017.2',
        'requests~=2.18.4',
        'six~=1.11.0',
    ],
    test_suite='runtests.runtests',
    tests_require=[
        'python-dotenv~=0.7.1',
        'dj-database-url~=0.4.2',
    ],
)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.4 on 2017-09-24 21:00
from __future__ import unicode_literals

from django.db import migrations
from django.db.models import Count


def forward_func(apps, schema_editor):
    """Forward direction: migrate data on Package to related models.

    Packages must have both a print run date and at least one assigned
    print section (also known as a "placement_type") for this migration
    to construct a new ContentPlacement model.
    """
    Package = apps.get_model('budget', 'Package')
    ContentPlacement = apps.get_model('budget', 'ContentPlacement')
    Publication = apps.get_model('budget', 'PrintPublication')

    placed_packages = Package.objects.annotate(
        num_placements=Count('print_section')
    ).exclude(
        num_placements=0
    ).filter(
        print_run_date__isnull=False
    )

    for package in placed_packages:
        print_section_id = list(set(
            package.print_section.values_list('publication__id', flat=True)
        ))

        print_sections_to_migrate = package.print_section.all()

        saved_publication_id = print_section_id[0]

        if len(print_section_id) != 1:
            print_sections_to_migrate = package.print_section.filter(
                publication__id=saved_publication_id
            )

            print(' '.join([
                'Multiple placement destinations found for budget.Package',
                'instance with an ID of {}.',
                'This migration will not create ContentPlacement models for',
                'the following destinations and placement types:'
            ]).format(package.id))

            for section in package.print_section.exclude(
                publication__id=saved_publication_id
            ):
                print('*   publication={}, type={}'.format(
                    section.publication.name,
                    section.name
                ))

            print(' '.join([
                'If the next migration to `budget` is allowed to run, these',
                'destination/placement-type pairs\' data will be deleted.',
            ]))

        published_destination = Publication.objects.get(
            id=saved_publication_id
        )

        # NOTE: This needs to be coerced to a basic list, to avoid
        # serialization issues during the migration.
        placement_type_slugs = list(print_sections_to_migrate.values_list(
            'slug',
            flat=True
        ))

        placement_obj = {
            'package': package,
            'destination': published_destination,
            'placement_types': placement_type_slugs,
            'placement_details': '',
            'run_date': package.print_run_date,
            'external_slug': package.print_system_slug,
            'is_finalized': package.is_print_placement_finalized,
        }

        ContentPlacement.objects.create(**placement_obj)


def backward_func(apps, schema_editor):
    """"""
    ContentPlacement = apps.get_model('budget', 'ContentPlacement')
    PrintSection = apps.get_model('budget', 'PrintSection')

    for placement in ContentPlacement.objects.all():
        if placement.package.print_section.count() == 0:
            pkg = placement.package

            pkg.print_run_date = placement.run_date
            pkg.print_system_slug = placement.external_slug
            pkg.is_print_placement_finalized = placement.is_finalized

            pkg.save()

            for placement_type in placement.placement_types:
                try:
                    type_model = PrintSection.objects.get(
                        publication__id=placement.destination.id,
                        slug=placement_type
                    )
                    pkg.print_section.add(type_model)
                    pkg.save()
                except PrintSection.DoesNotExist:
                    print(' '.join([
                        'Could not find a placement type with destination',
                        '"{}" and slug "{}" to attach to package with ID {}.',
                    ]).format(
                        placement.destination.name,
                        placement_type,
                        pkg.id
                    ))
        else:
            print(' '.join([
                'Couldn\'t save placement in destination "{}" of type(s) "{}"',
                'for package with ID {} because that package already has a',
                'stored placement.',
            ]).format(
                placement.destination.name,
                placement_type,
                pkg.id
            ))
            print('-   This placement is still saved with ID {}.'.format(
                placement.id
            ))


class Migration(migrations.Migration):

    dependencies = [
        ('budget', '0033_contentplacement'),
    ]

    operations = [
        migrations.RunPython(forward_func, backward_func)
    ]
